import asyncio
import datetime
import json

import aiohttp

from .. import fhem, generic


class energie_gv_at(generic.FhemModule):

    URL_PEAK_HOURS = "https://awareness.cloud.apg.at/api/v1/PeakHourStatus"
    URL_ELECTRICITY_NOW = (
        "https://energie.gv.at/api/gd/store/renewable_electricity_now_bars.json"
    )
    URL_RENEWABLE_NOW = (
        "https://energie.gv.at/api/gd/store/renewable_electricity_now_total.json"
    )

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 3:
            return "Usage: define energy_austria fhempy energie_gv_at"

        await self.set_icon("measure_photovoltaic_inst")

        self.create_async_task(self.update_loop())

    async def retrieve_data(self, url):
        # aiohttp get
        headers = {
            "accept": "application/json, text/plain, */*",
            "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
            "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        }
        try:
            async with aiohttp.ClientSession(headers=headers) as session:
                async with session.get(url) as resp:
                    if resp.status == 200:
                        r_json = json.loads(await resp.text())
                        return r_json
                    else:
                        await fhem.readingsSingleUpdate(
                            self.hash,
                            "state",
                            f"failed: HTTP error {resp.status}",
                            1,
                        )
                        self.logger.error(
                            f"Failed to fetch, failed with status {resp.status} {url}"
                        )
        except Exception as ex:
            self.logger.exception("Failed to update")
            await fhem.readingsSingleUpdate(
                self.hash,
                "state",
                f"failed: {ex}",
                1,
            )
        return None

    async def update_loop(self):
        while True:
            peak_hours = await self.retrieve_data(energie_gv_at.URL_PEAK_HOURS)
            await self.handle_peak_hours(peak_hours)
            elec_now = await self.retrieve_data(energie_gv_at.URL_ELECTRICITY_NOW)
            await self.handle_electricity_now(elec_now)
            renewable_now = await self.retrieve_data(energie_gv_at.URL_RENEWABLE_NOW)
            await self.handle_renewable_now(renewable_now)
            delta = datetime.timedelta(hours=1)
            now = datetime.datetime.now()
            next_hour = (now + delta).replace(microsecond=0, second=0, minute=0)
            wait_seconds = (next_hour - now).seconds
            await asyncio.sleep(wait_seconds + 10)

    async def handle_renewable_now(self, res):
        await fhem.readingsSingleUpdate(
            self.hash,
            f"current_hour_ratio_renewable",
            round(res["renewablePercentage"], 1),
            1,
        )
        await fhem.readingsSingleUpdate(
            self.hash,
            f"current_hour_ratio_conventional",
            round(res["conventionalPercentage"], 1),
            1,
        )

    async def handle_electricity_now(self, res):
        all_labels = [
            "Biomasse",
            "Erdgas",
            "Nettoimporte",
            "Sonne",
            "Sonstige",
            "Wasser",
            "Wind",
        ]
        for entry in res:
            await fhem.readingsSingleUpdate(
                self.hash,
                f"current_hour_electricity_{entry['label']}",
                round(entry["percentage"], 1),
                1,
            )
            all_labels.remove(entry["label"])
        for entry in all_labels:
            await fhem.readingsSingleUpdate(
                self.hash,
                f"current_hour_electricity_{entry}",
                0,
                1,
            )

    async def handle_peak_hours(self, res):
        # current_hour_category
        default_status = res["DefaultStatus"]
        curr_status = default_status
        status_infos = res["StatusInfos"]
        for entry in status_infos:
            utc = entry["utc"]
            entry_time = datetime.datetime.strptime(utc, "%Y-%m-%dT%H:%M:%S%z")
            entry_time_diff = (
                entry_time - datetime.datetime.now(datetime.timezone.utc)
            ).total_seconds()

            if 0 < entry_time_diff < 3599:
                curr_status = entry["s"]

        await fhem.readingsSingleUpdate(
            self.hash, "current_hour_category", curr_status, 1
        )

        state = "ok" if curr_status == "1" else "save energy"
        await fhem.readingsSingleUpdate(self.hash, "state", state, 1)



import asyncio
import functools

import aiohttp
from bs4 import BeautifulSoup

from .. import fhem, generic, utils


class fhem_forum(generic.FhemModule):

    URL_UNREADREPLIES = "https://forum.fhem.de/index.php?action=unreadreplies"
    URL_UNREAD = "https://forum.fhem.de/index.php?action=unread"
    URL_PM = "https://forum.fhem.de/index.php?action=pm"

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 15,
                "format": "int",
                "help": "Change interval in minutes, default is 15.",
            },
            "max_topics": {
                "default": 10,
                "format": "int",
                "help": "Maximum number of topics to show in readings.",
            },
            "keywords_unread_replies": {
                "default": "",
                "help": "Only topics with this keywords will be listed.",
                "function": "set_attr_keywords",
            },
            "keywords_unread": {
                "default": "",
                "help": "Only topics with this keywords will be listed.",
                "function": "set_attr_keywords",
            },
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define fhem.forum fhempy fhem_forum FHEM_COOKIE"

        self.cookie = args[3]
        self.first_run = True

        self.create_async_task(self.update_loop())

    async def set_attr_keywords(self, hash):
        self.create_async_task(self.update_once())

    async def update_once(self):
        stateset_ur = await self.get_fhem_data(fhem_forum.URL_UNREADREPLIES)
        stateset_u = await self.get_fhem_data(fhem_forum.URL_UNREAD)
        stateset_pm = await self.get_fhem_data(fhem_forum.URL_PM)
        if stateset_ur is False and stateset_u is False and stateset_pm is False:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "-", 1)

    async def update_loop(self):
        while True:
            # aiohttp get
            await self.update_once()
            await asyncio.sleep(self._attr_interval * 60)

    async def get_fhem_data(self, url):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    url,
                    headers={"cookie": "FHEM-Forum588=" + self.cookie},
                ) as resp:
                    if resp.status == 200:
                        return await self.handle_response(await resp.text(), url)
                    else:
                        await fhem.readingsSingleUpdate(
                            self.hash,
                            "state",
                            f"failed: HTTP error {resp.status}",
                            1,
                        )
                        self.logger.error(
                            f"Failed to fetch {self.update_url}, "
                            f"failed with status {resp.status}"
                        )
        except Exception:
            self.logger.exception("Failed to update")

    def contains_keyword(self, topic, keywords):
        if keywords == "":
            return True

        for kw in keywords.split(","):
            if topic.find(kw) > -1:
                return True
        return False

    def get_soup_entries(self, response):
        soup = BeautifulSoup(response, "html.parser")
        topics = soup.find("div", {"id": "topic_container"})
        if topics is None:
            return []
        entries = topics.findAll("div", {"class": "windowbg"})
        return entries

    def get_soup_private_messages(self, response):
        soup = BeautifulSoup(response, "html.parser")
        topics = soup.select(".windowbg.unread_pm")
        if topics is None:
            return []
        entries = soup.select(".windowbg.unread_pm")
        return entries

    async def handle_private_messages(self, response):
        stateset = False
        entries = await utils.run_blocking(
            functools.partial(self.get_soup_private_messages, response)
        )
        if len(entries) == 0:
            ret = await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"private_message", "-"
            )
            return stateset

        try:
            for entry in entries:
                title = entry.select(".pm_subject")[0].next.next.text
                user_from = entry.select(".pm_from_to")[0].text.strip()
                date_sent = entry.select(".pm_time")[0].text
                link_to_msg = entry.select(".pm_subject")[0].next.next["href"]
                link_to_msg = "https://forum.fhem.de/index.php?action=pm" + link_to_msg

                ret = await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    f"private_message",
                    f'<html><a href="{link_to_msg}" target="_blank">'
                    + f"{title}</a>"
                    + f"<br>{date_sent} von {user_from}</html>",
                )

                if ret:
                    stateset = True
                    ret = await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        f"state",
                        f'<html><a href="{link_to_msg}" target="_blank">'
                        + f"{title}</a>"
                        + f"<br>{date_sent} von {user_from}</html>",
                    )
                break

        except Exception:
            self.logger.exception("Failed to handle private messages")
        return stateset

    async def handle_response(self, response, url):
        # bs4
        stateset = False
        await fhem.readingsBeginUpdate(self.hash)
        try:
            reading = "topic_unread_replies"
            keywords = self._attr_keywords_unread_replies
            if url == fhem_forum.URL_UNREAD:
                reading = "topic_unread"
                keywords = self._attr_keywords_unread
            elif url == fhem_forum.URL_PM:
                await self.handle_private_messages(response)

            if url != fhem_forum.URL_PM:
                entries = await utils.run_blocking(
                    functools.partial(self.get_soup_entries, response)
                )

                i = 1
                for entry in entries:
                    subject = entry.find("div", {"class": "recent_title"})
                    if subject is None:
                        continue

                    if not self.contains_keyword(subject.find("span").text, keywords):
                        continue

                    link_to_new = subject.findAll("a")[0]["href"]
                    title_to_new = subject.find("span").find("a").text

                    last_post = entry.find("div", {"class": "lastpost"}).findAll("a")
                    last_post_date = last_post[0]
                    last_post_name = last_post[1]
                    last_post_date["target"] = "_blank"
                    last_post_name["target"] = "_blank"
                    last_post_str = f"{last_post_date} von {last_post_name}"

                    ret = await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        f"{reading}_{i:02d}",
                        f'<html><a href="{link_to_new}" target="_blank">'
                        + f"{title_to_new}</a>"
                        + f"<br>{last_post_str}</html>",
                    )

                    if i == 1 and (ret is not None or self.first_run is True):
                        self.first_run = False
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash,
                            "state",
                            f'<html><a href="{link_to_new}" target="_blank">'
                            + f"{title_to_new}</a>"
                            + f"<br>{last_post_str}</html>",
                        )

                    stateset = True

                    i += 1
                    if i > self._attr_max_topics:
                        break

                while i <= self._attr_max_topics:
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash, f"{reading}_{i:02d}", "-"
                    )
                    i += 1

        except Exception as ex:
            self.logger.exception(f"Failed to handle data from {url}")
            await fhem.readingsBulkUpdate(self.hash, "state", {ex}, 1)

        await fhem.readingsEndUpdate(self.hash, 1)

        return stateset



import asyncio

import aiohttp

from .. import fhem, generic


class aktionsfinder(generic.FhemModule):

    SEARCH_API = "https://api.post.at/aktionsfinder/afcore/api/public/v2/search"

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 24,
                "format": "int",
                "help": "Change interval, default is 24h.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) < 4:
            return "Usage: define water fhempy aktionsfinder voeslauer"

        self.search_item = " ".join(args[3:])

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        while True:
            # aiohttp post
            headers = {
                "authority": "api.post.at",
                "accept": "*/*",
                "accept-currency": "EUR",
                "accept-language": "de",
                "content-type": "application/json",
                "ocp-apim-subscription-key": "3b9e1fb43aa746e583919d322d2f239f",
                "origin": "https://www.aktionsfinder.at",
                "sec-ch-ua": 'Chromium";v="112", "Google Chrome";v="112", "Not:A-Brand";v="99"',
                "sec-ch-ua-mobile": "?0",
                "sec-ch-ua-platform": '"Chrome OS"',
                "sec-fetch-dest": "empty",
                "sec-fetch-mode": "cors",
                "sec-fetch-site": "cross-site",
                "user-agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36",
                "x-client-token": "YWZ3ZWJwd2E6OTNtMXBpZTY0bnN6NnVsZnRkeWQ0MjhuNjl6N296OHRwbXQwM3VkZQ==",
            }
            query_json = {
                "radius": "50000",
                "query": self.search_item,
                "lat": 48.21,
                "lng": 16.37,
            }
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.post(
                        aktionsfinder.SEARCH_API,
                        json=query_json,
                        headers=headers,
                    ) as resp:
                        if resp.status == 200:
                            await self.handle_response(await resp.json())
                        else:
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                "state",
                                f"failed: HTTP error {resp.status}",
                                1,
                            )
                            self.logger.error(f"Failed with status {resp.status}")
            except Exception:
                self.logger.exception("Failed to update")
            await asyncio.sleep(self._attr_interval * 3600)

    async def handle_response(self, response):
        await fhem.readingsBeginUpdate(self.hash)
        lowest_price = None
        for result in response["searchResult"]["content"]:
            if result["newPrice"] is None:
                continue

            if lowest_price is None or result["newPrice"] < lowest_price:
                lowest_price = result["newPrice"]
            else:
                # do not update if the price is not lowest
                continue

            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "vendor", result["vendorTitle"]
            )
            img_html = f"<html><img src=\"{result['image']['large']}\"/></html>"
            await fhem.readingsBulkUpdateIfChanged(self.hash, "image", img_html)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "valid_from", result["validFrom"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "valid_to", result["validTo"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "price_old", result["oldPrice"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "price_new", result["newPrice"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                "state",
                str(result["newPrice"])
                + "€ ("
                + result["validFrom"]
                + " - "
                + result["validTo"]
                + ")",
            )

        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio

from fhempy.lib import fhem, utils
from fhempy.lib.generic import FhemModule
from meross_iot.http_api import MerossHttpClient
from meross_iot.manager import MerossManager


class meross_setup:
    def __init__(self, logger, fhemdevice: FhemModule):
        self.logger = logger
        self.fhemdev = fhemdevice
        self.hash = fhemdevice.hash
        self.ready = False
        self._devices = []

    async def Define(self, hash, args, argsh):
        self._username = args[4]
        self._password = args[5]

        self.fhemdev.create_async_task(self.run_setup())

    async def run_setup(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "connecting", 1)

        # Setup the HTTP client API from user-password
        http_api_client = await MerossHttpClient.async_from_user_password(
            email=self._username, password=self._password
        )

        # Setup and start the device manager
        manager = MerossManager(http_client=http_api_client)
        await manager.async_init()

        # Retrieve all the MSS310 devices that are registered on this account
        await manager.async_device_discovery()
        self._devices = manager.find_devices()
        if len(self._devices) > 0:
            await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
            await self._init_devices()
            self.ready = True
        else:
            await fhem.readingsSingleUpdate(self.hash, "state", "failed to connect", 1)

    async def _create_fhem_device(self, name, device_id):
        devalias = name
        devname = name + "_" + device_id
        devname = utils.gen_fhemdev_name(devname)
        device_exists = await fhem.checkIfDeviceExists(
            self.hash, "FHEMPYTYPE", "meross", "DEVICEID", device_id
        )
        if not device_exists:
            self.logger.info(
                (f"create: {devname} fhempy meross " f"{self.hash['NAME']} {device_id}")
            )
            # define each device (CommandDefine ... tuya_cloud_setup_dev deviceid
            await fhem.CommandDefine(
                self.hash,
                (f"{devname} fhempy meross " f"{self.hash['NAME']} {device_id}"),
            )
            await fhem.CommandAttr(self.hash, f"{devname} alias {devalias}")
            # wait for FHEM to handle CommandDefine
            await asyncio.sleep(1)

    async def _init_devices(self):
        # wait for init to complete, otherwise devices might not be available yet
        while await fhem.init_done(self.hash) == 0:
            await asyncio.sleep(3)

        # iterate through device list
        for device in self._devices:
            await self._create_fhem_device(device.name, device.uuid)

    def get_device_by_id(self, id):
        for dev in self._devices:
            if dev.uuid == id:
                return dev
        return None



import asyncio
import re

import aiohttp
from bs4 import BeautifulSoup

from .. import fhem, generic


class ikos(generic.FhemModule):

    headers1 = {
        "authority": "ikosresorts.swapsystems.com",
        "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
        "cache-control": "max-age=0",
        "sec-ch-ua": '"Google Chrome";v="113", "Chromium";v="113", "Not-A.Brand";v="24"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"Chrome OS"',
        "sec-fetch-dest": "document",
        "sec-fetch-mode": "navigate",
        "sec-fetch-site": "none",
        "sec-fetch-user": "?1",
        "upgrade-insecure-requests": "1",
        "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
    }
    headers2 = {
        "accept": "*/*",
        "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
        "sec-ch-ua": '"Google Chrome";v="113", "Chromium";v="113", "Not-A.Brand";v="24"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"Chrome OS"',
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "x-requested-with": "XMLHttpRequest",
        "Referer": "https://ikosresorts.swapsystems.com/",
        "Referrer-Policy": "no-referrer-when-downgrade",
        "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36",
    }

    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        attr_config = {
            "interval": {
                "default": 24,
                "format": "int",
                "help": "Change interval in hours, default is 24 hours.",
            },
            "adults": {
                "default": 2,
                "format": "int",
                "help": "Number of adults in one room.",
            },
            "children": {
                "default": 0,
                "format": "int",
                "help": "Number of children in one room.",
            },
            "age_child1": {
                "default": 3,
                "format": "int",
                "help": "Age of first child.",
            },
            "age_child2": {
                "default": 3,
                "format": "int",
                "help": "Age of second child.",
            },
            "age_child3": {
                "default": 3,
                "format": "int",
                "help": "Age of third child.",
            },
            "age_child4": {
                "default": 3,
                "format": "int",
                "help": "Age of fourth child.",
            },
        }
        await self.set_attr_config(attr_config)

        if len(args) != 5:
            return "Usage: define june.holidays fhempy ikos [FROMDATE] [TODATE]"

        self.date_from = args[3]
        self.date_to = args[4]

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon weather_sun")

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        while True:
            # aiohttp post
            data = {
                "sid": "",
                "fls": "",
                "starRating": "",
                "boardType": "",
                "fCurID": "",
                "SortBy": "",
                "CityID": 0,
                "HotelID": 0,
                "Checkin": self.date_from,
                "Checkout": self.date_to,
                "Rooms": 1,
                "Adults1": self._attr_adults,
                "Children1": self._attr_children,
                "r1C1": self._attr_age_child1,
                "r1C2": self._attr_age_child2,
                "r1C3": self._attr_age_child3,
                "r1C4": self._attr_age_child4,
                "Adults2": 0,
                "Children2": 0,
                "r2C1": 0,
                "r2C2": 0,
                "r2C3": 0,
                "r2C4": 0,
                "Adults3": 0,
                "Children3": 0,
                "r3C1": 0,
                "r3C2": 0,
                "r3C3": 0,
                "r3C4": 0,
                "SortID": 1,
            }

            try:
                await fhem.readingsBeginUpdate(self.hash)
                async with aiohttp.ClientSession() as session:
                    async with session.get(
                        "https://ikosresorts.swapsystems.com",
                        headers=ikos.headers1,
                        json=data,
                    ) as resp:
                        pass
                    async with session.get(
                        "https://ikosresorts.swapsystems.com/hotel_results",
                        data=data,
                        headers=ikos.headers2,
                    ) as resp:
                        await self.handle_response(await resp.text())
            except Exception:
                self.logger.exception("Failed to update")

            await fhem.readingsEndUpdate(self.hash, 1)
            await asyncio.sleep(self._attr_interval * 60 * 60)

    async def handle_response(self, response):
        soup = BeautifulSoup(response, "html.parser")
        hotels = soup.findAll("div", {"class": "Search-header-rounded"})
        flights = soup.findAll("div", {"class": "flights"})
        best_price = None
        for i in range(0, len(hotels)):
            lowest_price_text = hotels[i].find("strong").text
            lowest_price = re.findall("\d+\.\d+", lowest_price_text)[0]
            hotel_name = hotels[i].find("a", {"class": "cardtitle htd"})["data-id2"]
            room_type = flights[i].find("a", {"data-toggle": "modal"})["data-id2"]
            if best_price is None or lowest_price < best_price:
                best_price = lowest_price
                best_hotel = hotel_name
                best_room = room_type
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"{i}_lowest_price", lowest_price
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"{i}_hotel_name", hotel_name
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, f"{i}_room_type", room_type
            )
        if best_price is None:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "state", f"No rooms available"
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(self.hash, "best_price", best_price)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "best_price_hotel", best_hotel
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "best_price_room", best_room
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "state", f"{best_hotel}: {best_price}€"
            )



import asyncio
import time

import aiohttp

from .. import fhem, generic


class websitetests(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 600,
                "format": "int",
                "help": "Change interval, default is 600s.",
            },
            "max_duration": {
                "default": 800,
                "format": "int",
                "help": (
                    "Error will be reported if duration_ms "
                    "higher than max_duration in ms."
                ),
            },
            "headers": {
                "default": {},
                "format": "json",
                "options": "textField-long",
                "help": (
                    "Example:<br>"
                    '{<br>"'
                    '"Content-type": "application/json",<br>'
                    '"Accept": "application/json, text/plain, */*",<br>'
                    '"Accept-Language": "en",<br>'
                    '"User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 15054.63.0) '
                    '"AppleWebKit/537.36 (KHTML, like Gecko) "'
                    'Chrome/106.0.0.0 Safari/537.36"<br>'
                    "}"
                ),
            },
            "response_contains": {
                "default": "",
                "options": "textField-long",
                "help": "Response must contain this text.",
            },
        }
        await self.set_attr_config(attr_config)

        if len(argsh) > 0:
            k = list(argsh)[0]
            url = k + "=" + argsh[k]
            args.append(url)
        if len(args) != 4:
            return "Usage: define webtest fhempy websitetests URL"

        self.update_url = args[3]

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        while True:
            # aiohttp get
            start_time = time.time()
            end_time = 0
            status = 0
            response_contains = -1
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(
                        self.update_url, headers=self._attr_headers
                    ) as resp:
                        status = resp.status
                        end_time = time.time()
                        await fhem.readingsSingleUpdate(
                            self.hash, "response_status", resp.status, 1
                        )
                        if resp.status == 200:
                            text = await resp.text()
                            if len(text) > 5000:
                                text = text[0:5000] + "..."
                            await fhem.readingsSingleUpdate(
                                self.hash, "response", text, 1
                            )
                            response_contains = text.find(self._attr_response_contains)
                            await fhem.readingsSingleUpdate(
                                self.hash, "response_contains", response_contains, 1
                            )

                        else:
                            await fhem.readingsSingleUpdate(
                                self.hash,
                                "state",
                                f"failed: HTTP error {resp.status}",
                                1,
                            )
                            self.logger.error(
                                f"Failed to fetch {self.update_url}, "
                                f"failed with status {resp.status}"
                            )
            except Exception:
                status = 0
                end_time = time.time()
                self.logger.exception("Failed to update")

            duration = (end_time - start_time) * 1000
            await fhem.readingsSingleUpdate(self.hash, "duration_ms", duration, 1)
            if status == 200:
                if duration < self._attr_max_duration:
                    await fhem.readingsSingleUpdate(self.hash, "state", "ok", 1)
                else:
                    await fhem.readingsSingleUpdate(
                        self.hash,
                        "state",
                        "duration too long: " + str(int(duration)),
                        1,
                    )
                if response_contains == -1:
                    await fhem.readingsSingleUpdate(
                        self.hash, "state", "incorrect response", 1
                    )

            await asyncio.sleep(self._attr_interval)



import asyncio

from .. import fhem, generic


class huawei_modbus(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.bridge = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {}
        await self.set_attr_config(attr_config)

        set_config = {}
        await self.set_set_config(set_config)

        if len(args) < 4:
            return "Usage: define my_sun2000 fhempy fusionsolar_modbus IP PORT SLAVE_ID"

        self.ip = args[3]
        self.port = 502
        self.slave_id = 1

        if len(args) >= 5:
            self.port = int(args[4])

            if len(args) == 6:
                self.slave_id = int(args[5])
            else:
                return (
                    "Usage: define my_sun2000 fhempy fusionsolar_modbus IP PORT"
                    + " SLAVE_ID"
                )

        self.hash["IP"] = self.ip
        self.hash["PORT"] = self.port
        self.hash["SLAVE_ID"] = self.slave_id

        self.create_async_task(self.start())

    async def Undefine(self, hash):
        if self.bridge:
            await self.bridge.stop()
        return await super().Undefine(self.hash)

    async def start(self):
        await self.connect()

        while True:
            try:
                await self.update()
            except Exception as e:
                self.logger.error(e)

            await asyncio.sleep(30)

    async def connect(self):
        from huawei_solar import HuaweiSolarBridge

        self.bridge = await HuaweiSolarBridge.create(
            host=self.ip, port=self.port, slave_id=self.slave_id
        )

    async def update(self):
        data = await self.bridge.update()

        await fhem.readingsBeginUpdate(self.hash)
        try:
            for key in data:
                await fhem.readingsBulkUpdateIfChanged(self.hash, key, data[key].value)
        except Exception as e:
            self.logger.error(e)
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import json
import logging
import os
import platform
import random
import socket
import time
import traceback
from datetime import datetime

import aiohttp
import websockets

from .version import __version__

logger = logging.getLogger(__name__)

function_active = []
update_locks = {}
wsconnection = None

# TODO use run_coroutine_threadsafe if asyncio.get_event_loop() == None
# this would make all functions threadsafe


def updateConnection(ws):
    global wsconnection
    wsconnection = ws


def setFunctionActive(hash):
    function_active.append(hash["NAME"])


def setFunctionInactive(hash):
    element = function_active.pop()
    if element != hash["NAME"]:
        logger.error(
            f"Set wrong function inactive, tried {hash['NAME']}, "
            f"current function_active: {function_active},{element}"
        )


async def getDeviceHashName(hash, typeinternal, typevalue, internal, value):
    cmd = (
        "foreach my $fhem_dev (sort keys %main::defs) {"
        + "  return $main::defs($fhem_dev}{NAME}) if(defined($main::defs{$fhem_dev}{"
        + typeinternal
        + "}) && $main::defs{$fhem_dev}{"
        + typeinternal
        + "} eq '"
        + typevalue
        + "' && $main::defs{$fhem_dev}{"
        + internal
        + "} eq '"
        + value
        + "');;"
        + "}"
        + "return 0;;"
    )
    return await sendCommandHash(hash, cmd)


async def getUniqueId(hash):
    cmd = "getUniqueId()"
    return await sendCommandHash(hash, cmd)


async def init_done(hash):
    cmd = "$init_done"
    res = await sendCommandHash(hash, cmd)
    return int(res)


async def ReadingsVal(name, reading, default):
    cmd = "ReadingsVal('" + name + "', '" + reading + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def AttrVal(name, attr, default):
    cmd = "AttrVal('" + name + "', '" + attr + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def InternalVal(name, internal, default):
    cmd = "InternalVal('" + name + "', '" + internal + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def addToDevAttrList(name, attr_list):
    cmd = "addToDevAttrList('" + name + "', '" + attr_list + "')"
    return await sendCommandName(name, cmd)


async def setDevAttrList(name, attr_list):
    attr_list += " IODev disable:0,1"
    cmd = "setDevAttrList('" + name + "', '" + attr_list + " '.$readingFnAttributes)"
    return await sendCommandName(name, cmd)


async def readingsBeginUpdate(hash):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    try:
        await asyncio.wait_for(update_locks[hash["NAME"]].acquire(), 120)
    except asyncio.TimeoutError:
        logging.error(
            f"{hash['NAME']}: readingsBeginUpdate couldn't acquire lock,"
            + " caused by readingsBeginUpdate without End or Single update inbetween"
        )
    cmd = "readingsBeginUpdate($defs{'" + hash["NAME"] + "'});;"
    return await sendCommandHash(hash, cmd)


async def readingsBulkUpdateIfChanged(hash, reading, value):
    try:
        value = convertValue(value)
        cmd = (
            "readingsBulkUpdateIfChanged($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "');;"
        )
        if hash["NAME"] not in update_locks or not update_locks[hash["NAME"]].locked():
            logging.error(
                f"{hash['NAME']}: readingsBulkUpdateIfChanged without "
                + f"readingsBeginUpdate: {cmd}"
            )
            return
        return await sendCommandHash(hash, cmd)
    except Exception:
        logger.exception("Failed to do readingsBulkUpdateIfChanged")


async def readingsBulkUpdate(hash, reading, value, changed=None):
    try:
        value = convertValue(value)
        if changed is None:
            cmd = (
                "readingsBulkUpdate($defs{'"
                + hash["NAME"]
                + "'},'"
                + reading
                + "','"
                + value.replace("'", "\\'")
                + "');;"
            )
        else:
            cmd = (
                "readingsBulkUpdate($defs{'"
                + hash["NAME"]
                + "'},'"
                + reading
                + "','"
                + value.replace("'", "\\'")
                + "', "
                + str(changed)
                + ");;"
            )
        if hash["NAME"] not in update_locks or not update_locks[hash["NAME"]].locked():
            logging.error(
                f"{hash['NAME']}: readingsBulkUpdate without "
                + f"readingsBeginUpdate: {cmd}"
            )
            return
        return await sendCommandHash(hash, cmd)
    except Exception:
        logger.exception("Failed to do readingsBulkUpdate")


async def readingsEndUpdate(hash, do_trigger):
    if hash["NAME"] not in update_locks:
        logger.error("readingsEndUpdate without active readingsBeginUpdate")
    cmd = "readingsEndUpdate($defs{'" + hash["NAME"] + "'}," + str(do_trigger) + ");;"
    res = await sendCommandHash(hash, cmd)
    update_locks[hash["NAME"]].release()
    return res


async def readingsSingleUpdate(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = (
            "readingsSingleUpdate($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "',"
            + str(do_trigger)
            + ")"
        )
        return await sendCommandHash(hash, cmd)


async def readingsSingleUpdateIfChanged(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = (
            "readingsBeginUpdate($defs{'"
            + hash["NAME"]
            + "'});;readingsBulkUpdateIfChanged($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "');;readingsEndUpdate($defs{'"
            + hash["NAME"]
            + "'},"
            + str(do_trigger)
            + ");;"
        )
        return await sendCommandHash(hash, cmd)


async def CommandDefine(hash, definition: str):
    cmd = 'CommandDefine(undef, "' + definition + '")'
    ret = await sendCommandHash(hash, cmd)
    if ret is not None:
        return ret

    pos_fhempy = definition.find(" fhempy ")
    if pos_fhempy == -1:
        pos_fhempy = definition.find(" PythonModule ")
    if pos_fhempy > 0:
        devname = definition.split(" ")[0]
        iodev = await AttrVal(hash["NAME"], "IODev", "")
        if iodev != "":
            await CommandAttr(hash, f"{devname} IODev {iodev}")

        fhempytype = definition.split(" ")[2]
        await CommandAttr(hash, f"{devname} group {fhempytype}")

        room = await AttrVal(hash["NAME"], "room", "")
        if room != "":
            await CommandAttr(hash, f"{devname} room fhempy")


async def CommandAttr(hash, attrdef):
    cmd = 'CommandAttr(undef, "' + attrdef.replace('"', '\\"') + '")'
    return await sendCommandHash(hash, cmd)


async def CommandDeleteReading(hash, deldef):
    cmd = 'CommandDeleteReading(undef, "' + deldef + '")'
    return await sendCommandHash(hash, cmd)


async def checkIfDeviceExists(hash, typeinternal, typevalue, internal, value):
    cmd = (
        "foreach my $fhem_dev (sort keys %main::defs) {"
        + "  return 1 if(defined($main::defs{$fhem_dev}{"
        + typeinternal
        + "}) && $main::defs{$fhem_dev}{"
        + typeinternal
        + "} eq '"
        + typevalue
        + "' && defined($main::defs{$fhem_dev}{"
        + internal
        + "}) && $main::defs{$fhem_dev}{"
        + internal
        + "} eq '"
        + value
        + "');;"
        + "}"
        + "return 0;;"
    )
    return await sendCommandHash(hash, cmd)


# UTILS FUNCTIONS TO SEND COMMAND TO FHEM
def convertValue(value):
    if value is None:
        value = ""
    if value is True:
        value = 1
    if value is False:
        value = 0
    if isinstance(value, datetime):
        value = value.strftime("%Y-%m-%d %H:%M:%S")

    return str(value)


async def get_github_data():
    res_json = {}
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://api.github.com/repos/fhempy/fhempy/releases/latest"
            ) as resp:
                res_json = await resp.json()
    except Exception:
        logger.exception("Failed to get github fhempy data")
    return res_json


async def send_latest_release():
    while True:
        try:
            github_data = await get_github_data()
            try:
                latest_version = github_data["name"][1:]
            except Exception:
                latest_version = "unknown"

            msg = {
                "msgtype": "version",
                "version_available": latest_version,
                "version_release_notes": (
                    '<html><a href="https://github.com/fhempy/fhempy/releases" target="_blank">'
                    "Release Notes</a></html>"
                ),
            }
            msg = json.dumps(msg, ensure_ascii=False)
            logger.debug("<<< WS: " + msg)
            global wsconnection
            await wsconnection.send(msg)
        except Exception:
            logger.exception("Failed to update latest release infos")
        await asyncio.sleep(3600 * 12)


async def send_version():
    msg = {
        "msgtype": "version",
        "version": __version__,
        "python": platform.python_version(),
        "os": os.name,
        "system": platform.system(),
        "release": platform.release(),
        "hostname": socket.gethostname(),
    }
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    global wsconnection
    await wsconnection.send(msg)


async def send_default_response(hash, set_default_response):
    msg = {
        "msgtype": "set_update",
        "NAME": hash["NAME"],
        "set_default_response": set_default_response,
    }
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    global wsconnection
    await wsconnection.send(msg)


async def send_and_wait(name, cmd):
    fut = asyncio.get_running_loop().create_future()
    msg = {
        "awaitId": random.randint(10000000, 99999999),
        "NAME": name,
        "msgtype": "command",
        "command": cmd,
    }
    sent_time = time.time()

    def listener(rmsg):
        try:
            recv_time = time.time()
            fhem_time = (recv_time - sent_time) * 1000
            if fhem_time > 5000:
                # log error message if fhem took too long to handle cmd
                logger.error(f"FHEM took {fhem_time:.0f}ms for {cmd}")
            rmsg = json.loads(rmsg)
            logger.debug(f">>> {rmsg['awaitId']:08d} {fhem_time:.2f}ms: {rmsg}")
            fut.set_result(rmsg)
        except Exception:
            logger.error(f"Failed to set result, received: {rmsg}")

    global wsconnection
    wsconnection.register_msg_listener(listener, msg["awaitId"])
    logger.debug(f"<<< {msg['awaitId']:08d}: {msg}")
    msg = json.dumps(msg, ensure_ascii=False)
    try:
        await wsconnection.send(msg)
    except websockets.exceptions.ConnectionClosed:
        logger.error("Connection closed, can't send message.")
    except Exception as e:
        logger.error("Failed to send message via websocket: " + e)
        fut.set_exception(Exception("Failed to send message via websocket"))

    return await fut


async def sendCommandName(name, cmd, hash=None):
    ret = ""
    timeout = 180
    try:
        start = time.time()
        while len(function_active) != 0:
            if function_active[-1] == name:
                break
            await asyncio.sleep(0.1)
        end = time.time()
        duration = end - start
        if duration > 5:
            logger.error(f"sendCommandName took {duration}s to send: {cmd}")
        # wait max 180s for reply from FHEM
        jsonmsg = await asyncio.wait_for(send_and_wait(name, cmd), timeout)
        ret = jsonmsg["result"]
    except asyncio.TimeoutError:
        logger.error(f"NO RESPONSE since {timeout}s: " + cmd)
        ret = ""
    except asyncio.CancelledError:
        # task was cancelled
        pass
    except Exception as e:
        logger.error("Exception while waiting for reply: " + e)
        traceback.format_exc()
        ret = str(e)

    return ret


async def sendCommandHash(hash, cmd):
    return await sendCommandName(hash["NAME"], cmd, hash)



from __future__ import annotations

import asyncio
import datetime

from httpx import AsyncClient
from vaillant_netatmo_api import (
    AuthClient,
    SetpointMode,
    SystemMode,
    ThermostatClient,
    TokenStore,
)

from .. import fhem, generic, utils


class erelax_vaillant(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        unknown_key = "thisisnotmagic012345678901234567"
        self.unknown_code = utils.decrypt_string(
            (
                "eJwVwdsOgiAAANAvalNLrEcJM5chh"
                "RvOF6dZaBcCLBG/vnUOD/8gr1RNJ"
                "iNsgTOZrebk5njuM7Es\nlo4SBja"
                "S+yoZ+YHh61CIsF/qvE6Fsmr30sh"
                "FXeCt+WNoiAZ9R++EGRmXZ5pH+zh"
                "tWbjxzZa+JwjQ\nFGFEmvJ7vFgI3"
                "NYLPgs869OIiU9/UCcvrg==\n"
            ),
            unknown_key,
        )

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        attr_config = {
            "update_interval": {
                "default": 300,
                "format": "int",
                "help": "Change interval, default is 300.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "away": {},
            "home": {},
            "desiredTemperature": {
                "args": ["temperature"],
                "params": {"temperature": {"format": "float"}},
                "options": "slider,7,0.5,30",
            },
            "desiredTemperatureDuration": {
                "args": ["temperature", "duration"],
                "params": {
                    "duration": {"format": "int"},
                    "temperature": {"format": "float"},
                },
                "help": (
                    "Set temperature for a specific duration in minutes<br>"
                    "set vaillant TEMPERATURE DURATION"
                ),
            },
            "system_mode": {
                "args": ["mode"],
                "argsh": ["mode"],
                "options": "winter,summer,frostguard",
            },
        }
        await self.set_set_config(set_config)
        await self.set_icon("sani_heating_temp")
        if len(args) < 4:
            return "Usage: define erelax_vaillant fhempy USERNAME PASSWORD [STATION]"
        await fhem.readingsSingleUpdate(hash, "state", "connecting", 1)
        self.username = args[3]
        self.password = args[4]
        self.create_async_task(self.do_update_loop())

    async def set_away(self, hash, params):
        self.create_async_task(
            self.thermostat_client.async_set_minor_mode(
                self.devices[0].id,
                self.devices[0].modules[0].id,
                SetpointMode.AWAY,
                True,
                None,
                None,
            )
        )

    async def set_home(self, hash, params):
        self.create_async_task(
            self.thermostat_client.async_set_minor_mode(
                self.devices[0].id,
                self.devices[0].modules[0].id,
                SetpointMode.HWB,
                True,
                None,
                None,
            )
        )

    async def set_desiredTemperature(self, hash, params):
        endtime = datetime.datetime.now() + datetime.timedelta(
            minutes=self.devices[0].setpoint_default_duration
        )
        self.create_async_task(
            self.thermostat_client.async_set_minor_mode(
                self.devices[0].id,
                self.devices[0].modules[0].id,
                SetpointMode.MANUAL,
                True,
                endtime,
                params["temperature"],
            )
        )

    async def set_desiredTemperatureDuration(self, hash, params):
        endtime = datetime.datetime.now() + datetime.timedelta(
            minutes=params["duration"]
        )
        self.create_async_task(
            self.thermostat_client.async_set_minor_mode(
                self.devices[0].id,
                self.devices[0].modules[0].id,
                SetpointMode.MANUAL,
                True,
                endtime,
                params["temperature"],
            )
        )

    async def set_system_mode(self, hash, params):
        if params["mode"] == "winter":
            set_mode = SystemMode.WINTER
        elif params["mode"] == "summer":
            set_mode = SystemMode.SUMMER
        else:
            set_mode = SystemMode.FROSTGUARD
        self.create_async_task(
            self.thermostat_client.async_set_system_mode(
                self.devices[0].id, self.devices[0].modules[0].id, set_mode
            )
        )

    async def do_update_loop(self):
        async with AsyncClient() as client:
            token_store = TokenStore(
                "na_client_android_vaillant",
                self.unknown_code,
                None,
                None,
            )
            await self.async_get_token(client, token_store)
            self.thermostat_client = ThermostatClient(client, token_store)

            while True:
                try:
                    self.devices = (
                        await self.thermostat_client.async_get_thermostats_data()
                    )

                    await self.update_readings()
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "connected", 1
                    )
                except Exception as ex:
                    self.logger.exception(ex)
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "error", 1
                    )

                await asyncio.sleep(self._attr_update_interval)

    async def async_get_token(self, client, token_store):
        client = AuthClient(client, token_store)

        await client.async_token(self.username, self.password, "vaillant", "1.0.4.0")

    async def update_readings(self):
        if len(self.devices) == 0:
            await fhem.readingsSingleUpdate(
                self.hash, "state", "no stations found for this account", 1
            )
            return

        # currently only one thermostat (module) is supported
        v_station = self.devices[0]
        v_module = self.devices[0].modules[0]
        await fhem.readingsBeginUpdate(self.hash)

        # station
        await fhem.readingsBulkUpdate(self.hash, "station_name", v_station.station_name)
        await fhem.readingsBulkUpdate(self.hash, "system_mode", v_station.system_mode)
        await fhem.readingsBulkUpdate(self.hash, "station_mac", v_station.id)
        await fhem.readingsBulkUpdate(self.hash, "station_type", v_station.type)
        await fhem.readingsBulkUpdate(self.hash, "station_firmware", v_station.firmware)
        await fhem.readingsBulkUpdate(
            self.hash,
            "station_setpoint_default_duration",
            v_station.setpoint_default_duration,
        )
        await fhem.readingsBulkUpdate(
            self.hash,
            "station_setpoint_hwb",
            v_station.setpoint_hwb.setpoint_activate,
        )

        # module
        await fhem.readingsBulkUpdate(self.hash, "thermostat_mac", v_module.id)
        await fhem.readingsBulkUpdate(self.hash, "thermostat_type", v_module.type)
        await fhem.readingsBulkUpdate(
            self.hash, "thermostat_name", v_module.module_name
        )
        await fhem.readingsBulkUpdate(
            self.hash, "thermostat_firmware", v_module.firmware
        )
        await fhem.readingsBulkUpdate(
            self.hash, "thermostat_battery_percent", v_module.battery_percent
        )
        await fhem.readingsBulkUpdate(
            self.hash, "thermostat_boiler_status", v_module.boiler_status
        )
        await fhem.readingsBulkUpdate(self.hash, "thermostat_id", v_module.id)

        # setpoint_manual
        await fhem.readingsBulkUpdate(
            self.hash, "setpoint_manual", v_module.setpoint_manual.setpoint_activate
        )
        if v_module.setpoint_manual.setpoint_activate:
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_manual_endtime",
                v_module.setpoint_manual.setpoint_endtime,
            )
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_manual_endtime_epoch",
                v_module.setpoint_manual.setpoint_endtime.timestamp(),
            )
        else:
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_manual_endtime",
                "0",
            )

        # setpoint_away
        await fhem.readingsBulkUpdate(
            self.hash, "setpoint_away", v_module.setpoint_away.setpoint_activate
        )
        if v_module.setpoint_away.setpoint_activate:
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_away_endtime",
                v_module.setpoint_away.setpoint_endtime,
            )
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_away_endtime_epoch",
                v_module.setpoint_away.setpoint_endtime.timestamp(),
            )
        else:
            await fhem.readingsBulkUpdate(
                self.hash,
                "setpoint_away_endtime",
                "0",
            )

        # measured
        await fhem.readingsBulkUpdate(
            self.hash, "temperature", v_module.measured.temperature
        )
        await fhem.readingsBulkUpdate(
            self.hash, "est_setpoint_temp", v_module.measured.est_setpoint_temp
        )
        await fhem.readingsBulkUpdate(
            self.hash, "desiredTemperature", v_module.measured.setpoint_temp
        )
        await fhem.readingsBulkUpdate(
            self.hash, "outdoor_temperature", v_station.outdoor_temperature.te
        )
        await fhem.readingsBulkUpdate(
            self.hash,
            "outdoor_temperature_updated",
            v_station.outdoor_temperature.ti,
        )
        await fhem.readingsBulkUpdate(self.hash, "dhw", v_station.dhw)
        await fhem.readingsBulkUpdate(self.hash, "dhw_min", v_station.dhw_min)
        await fhem.readingsBulkUpdate(self.hash, "dhw_max", v_station.dhw_max)
        await fhem.readingsBulkUpdate(self.hash, "wifi_status", v_station.wifi_status)

        await fhem.readingsBulkUpdate(self.hash, "rf_status", v_module.rf_status)

        """ await fhem.readingsBulkUpdate(
            self.hash, "place_altitude", v_station.place_altitude
        )
        await fhem.readingsBulkUpdate(self.hash, "place_city", v_station.place_city)
        await fhem.readingsBulkUpdate(
            self.hash, "place_continent", v_station.place_continent
        )
        await fhem.readingsBulkUpdate(
            self.hash, "place_country_name", v_station.place_country_name
        )
        await fhem.readingsBulkUpdate(self.hash, "place_street", v_station.place_street)
        await fhem.readingsBulkUpdate(
            self.hash, "place_timezone", v_station.place_timezone
        )
        await fhem.readingsBulkUpdate(
            self.hash, "place_location", v_station.place_location
        )
        await fhem.readingsBulkUpdate(self.hash, "oem_serial", v_station.oem_serial)
        await fhem.readingsBulkUpdate(
            self.hash, "hot_water_anticipating", v_station.hot_water_anticipating
        )
        await fhem.readingsBulkUpdate(
            self.hash, "boiler_oem_serial", v_station.boiler_oem_serial
        )
        await fhem.readingsBulkUpdate(
            self.hash, "refill_water", v_station.ebus_refill_water
        )

        """
        await fhem.readingsEndUpdate(self.hash, 1)
